"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from flask_login import current_user
from puffotter.flask.test.TestFramework import _TestFramework


class TestLoginRoute(_TestFramework):
    """
    Class that tests log-in features
    """

    def test_page_get(self):
        """
        Tests getting the page
        :return: None
        """
        resp = self.client.get("/login")
        self.assertTrue(b"<!--user_management/login.html-->" in resp.data)

    def test_logging_in_and_out(self):
        """
        Tests logging in successfully, then once more, then logging out
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        with self.client:
            self.assertFalse(bool(current_user))
            resp = self.client.post("/login", follow_redirects=True, data={
                "username": user.username,
                "password": password
            })
            self.assertTrue(b"<!--static/index.html-->" in resp.data)
            self.assertTrue(b"Logged in successfully" in resp.data)
            self.assertTrue(current_user.is_authenticated)

            resp = self.client.get("/profile")
            self.assertTrue(user.username.encode("utf-8") in resp.data)

            resp = self.client.post("/login", follow_redirects=True, data={
                "username": user.username,
                "password": password
            })

            logged_in = self.config.STRINGS["logged_in"]
            already_logged_in = self.config.STRINGS["user_already_logged_in"]
            logged_out = self.config.STRINGS["logged_out"]

            self.assertFalse(logged_in.encode("utf-8") in resp.data)
            self.assertTrue(already_logged_in.encode("utf-8") in resp.data)
            self.assertTrue(b"<!--user_management/login.html-->" in resp.data)
            self.assertTrue(current_user.is_authenticated)

            resp = self.client.get("/logout", follow_redirects=True)
            self.assertFalse(current_user.is_authenticated)
            self.assertTrue(logged_out.encode("utf-8") in resp.data)
            self.assertTrue(b"<!--static/index.html-->" in resp.data)

    def test_invalid_login_attempts(self):
        """
        Tests trying to log in with invalid credentials etc
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        user2, password2, _ = self.generate_sample_user(confirmed=False)
        with self.client:
            for params in [
                {
                    "username": user.username,
                    "password": password + "AAA",
                    "expected": "invalid_password"
                },
                {
                    "username": user.username + "AAA",
                    "password": password,
                    "expected": "user_does_not_exist"
                },
                {
                    "username": user2.username,
                    "password": password2,
                    "expected": "user_is_not_confirmed"
                }
            ]:
                resp = self.client.post(
                    "/login", follow_redirects=True, data=params
                )
                logged_in = self.config.STRINGS["logged_in"]
                expected = self.config.STRINGS[params["expected"]]
                print(expected)
                print(resp.data)

                self.assertFalse(logged_in.encode("utf-8") in resp.data)
                self.assertTrue(expected.encode("utf-8") in resp.data)
                self.assertTrue(
                    b"<!--user_management/login.html-->" in resp.data
                )
                self.assertFalse(current_user.is_authenticated)

                resp = self.client.get("/profile")
                self.assertFalse(user.username.encode("utf-8") in resp.data)



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fat-ffipd.

fat-ffipd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fat-ffipd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fat-ffipd.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Union
from flask import render_template, Blueprint
from werkzeug import Response

static_blueprint = Blueprint("static", __name__)


@static_blueprint.route("/", methods=["GET"])
def index() -> Union[Response, str]:
    """
    The index page
    :return: The response
    """
    return render_template("static/index.html")


@static_blueprint.route("/about", methods=["GET"])
def about() -> Union[Response, str]:
    """
    The about page/"Impressum" for the website
    :return: The response
    """
    return render_template("static/about.html")


@static_blueprint.route("/privacy", methods=["GET"])
def privacy() -> Union[Response, str]:
    """
    Page containing a privacy disclaimer
    :return: The response
    """
    return render_template("static/privacy.html")



"""LICENSE
Copyright 2017 Hermann Krumrey <hermann@krumreyh.com>

This file is part of bundesliga-tippspiel.

bundesliga-tippspiel is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

bundesliga-tippspiel is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with bundesliga-tippspiel.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any
from puffotter.flask.base import db
from puffotter.flask.db.ModelMixin import ModelMixin
from puffotter.crypto import verify_password


class User(ModelMixin, db.Model):
    """
    Model that describes the 'users' SQL table
    A User stores a user's information, including their email address, username
    and password hash
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    __tablename__ = "users"
    """
    The name of the table
    """

    username = db.Column(db.String(12), nullable=False, unique=True)
    """
    The user's username
    """

    email = db.Column(db.String(150), nullable=False, unique=True)
    """
    The user's email address
    """

    password_hash = db.Column(db.String(255), nullable=False)
    """
    The user's hashed password, salted and hashed.
    """

    confirmed = db.Column(db.Boolean, nullable=False, default=False)
    """
    The account's confirmation status. Logins should be impossible as long as
    this value is False.
    """

    confirmation_hash = db.Column(db.String(255), nullable=False)
    """
    The account's confirmation hash. This is the hash of a key emailed to
    the user. Only once the user follows the link in the email containing the
    key will their account be activated
    """

    @property
    def is_authenticated(self) -> bool:
        """
        Property required by flask-login
        :return: True if the user is confirmed, False otherwise
        """
        return True

    @property
    def is_anonymous(self) -> bool:
        """
        Property required by flask-login
        :return: True if the user is not confirmed, False otherwise
        """
        return not self.is_authenticated  # pragma: no cover

    @property
    def is_active(self) -> bool:
        """
        Property required by flask-login
        :return: True
        """
        return self.confirmed

    def get_id(self) -> str:
        """
        Method required by flask-login
        :return: The user's ID as a unicode string
        """
        return str(self.id)

    def verify_password(self, password: str) -> bool:
        """
        Verifies a password against the password hash
        :param password: The password to check
        :return: True if the password matches, False otherwise
        """
        return verify_password(password, self.password_hash)

    def verify_confirmation(self, confirmation_key: str) -> bool:
        """
        Verifies a confirmation key against the confirmation hash
        :param confirmation_key: The key to check
        :return: True if the key matches, False otherwise
        """
        return verify_password(confirmation_key, self.confirmation_hash)

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models will be
                                 included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        data = {
            "id": self.id,
            "username": self.username,
            "email": self.email,
            "confirmed": self.confirmed
        }
        if include_children:
            pass
        return data



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import List, Any, Callable, Optional, Set
from colorama import Fore, Style


def yn_prompt(
        message: str,
        make_sure: bool = True,
        case_sensitive: bool = False
) -> bool:
    """
    Creates a yes/no prompt
    :param message: The message to display
    :param make_sure: Continuously prompts if the response is neither
                      'y' or 'n' until it is.
                      If false, every input besides 'y' will result in the
                      return value being False
    :param case_sensitive: Whether or not the prompt should be case-sensitive
    :return: True if the user specified 'y', else False
    """
    resp = input("{} {}(y|n){}: ".format(
        message, Fore.LIGHTYELLOW_EX, Style.RESET_ALL
    )).strip()
    if not case_sensitive:
        resp = resp.lower()

    if resp == "y":
        return True
    elif resp == "n" or not make_sure:
        return False
    else:
        return yn_prompt(message, make_sure, case_sensitive)


def selection_prompt(objects: List[object]) -> List[object]:
    """
    Prompts the user for a selection from a list of objects
    :param objects: The objects to show
    :return: The selection of objects
    """
    fill = len(str(len(objects)))
    for i, obj in enumerate(objects):
        print("[{}] {}".format(str(i + 1).zfill(fill), str(obj)))

    while True:

        selection = input("Selection: ").split(",")
        valid = True
        selected_objects = []

        for item in selection:

            try:
                start_index = int(item)
                end_index = start_index
            except ValueError:
                try:
                    start, end = item.split("-", 1)
                    start_index = int(start)
                    end_index = int(end)
                except (IndexError, ValueError):
                    valid = False
                    break

            for i in range(start_index, end_index + 1):
                try:
                    selected_objects.append(objects[i - 1])
                except IndexError:
                    valid = False
                    break

        if not valid:
            print("Invalid selection")
        else:
            return selected_objects


def prompt_comma_list(
        message: str,
        primitive_type: Callable[[str], Any] = str,
        min_count: int = 0,
        no_empty: bool = True,
        default: Optional[List[Any]] = None
) -> List[Any]:
    """
    Prompts the user for a comma-separated list
    :param message: The message to display
    :param primitive_type: The primitive type of the elements in the list
    :param min_count: The minimum amount of elements to be provided by the user
    :param no_empty: Removes any empty strings
    :param default: A default value
    :return: The result of the prompt
    """
    while True:
        try:
            prompt_message = message
            if default is not None:
                prompt_message = "{} {}{}{}".format(
                    prompt_message,
                    Fore.LIGHTGREEN_EX,
                    default,
                    Style.RESET_ALL
                )
            prompt_message += ": "

            response = input(prompt_message).strip()

            if default is not None and response == "":
                return default

            if default is not None and response == "[]":
                result = []  # type: List[str]
            else:
                result = list(map(lambda x: x.strip(), response.split(",")))

            if "" in result and no_empty:
                result.remove("")
            result = list(map(lambda x: primitive_type(x), result))

            if len(result) < min_count:
                print("Not enough values")
                raise ValueError()
            else:
                return result

        except (ValueError, TypeError):
            print("Invalid input")


def prompt(
        prompt_text: str = "",
        default: Optional[Any] = None,
        _type: Callable[[str], Any] = str,
        required: bool = True,
        choices: Optional[Set[str]] = None
) -> Optional[Any]:
    """
    Generic prompt with configuration options
    :param prompt_text: The text to display before the prompt
    :param default: A default value to use if the user responds with ''
    :param _type: The type of the object prompted. Must take a single string
                  as a parameter
    :param required: Whether or not a response is required
    :param choices: Valid choices for the prompt
    :return: The prompt result. May be None if required is False
    """
    prompt_message = prompt_text

    if choices is not None:
        prompt_message = "{} {}{}{}".format(
            prompt_message,
            Fore.LIGHTYELLOW_EX,
            choices,
            Style.RESET_ALL
        )

    if default is not None:
        prompt_message = "{} {}{}{}".format(
            prompt_message,
            Fore.LIGHTGREEN_EX,
            default,
            Style.RESET_ALL
        )

    prompt_message += ": "

    response = input(prompt_message).strip()

    while \
            response == "" and default is None \
            or (
                    choices is not None
                    and response not in choices
                    and default is None
            ):
        response = input(prompt_message).strip()

    if response == "" and default is not None:
        return default
    elif response == "":
        if required:
            return prompt(prompt_text, default, _type, required)
        else:
            return None
    else:
        try:
            return _type(response)
        except (TypeError, ValueError):
            return prompt(prompt_text, default, _type, required)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
import shutil
from base64 import b64encode
from unittest import TestCase
from typing import Tuple, Dict, List
from flask.blueprints import Blueprint
from puffotter.crypto import generate_random, generate_hash
from puffotter.flask.Config import Config
from puffotter.flask.db.User import User
from puffotter.flask.db.ApiKey import ApiKey
from puffotter.flask.initialize import init_flask, app, db
from puffotter.env import load_env_file


class _TestFramework(TestCase):
    """
    Class that models a testing framework for the flask application
    """

    module_name: str = "puffotter"
    models: List[db.Model] = []
    blueprints: List[Blueprint] = []
    root_path: str = "."

    def setUp(self):
        """
        Sets up the flask application and a temporary database to test with
        :return: None
        """
        self.db_path = ""
        self.temp_templates_dir = "templates"
        self.templates_sample_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "templates"
        )
        os.environ["FLASK_TESTING"] = "1"
        os.environ["FLASK_SECRET"] = generate_random(20)

        if self.module_name == "puffotter":
            if os.path.isdir(self.temp_templates_dir):
                shutil.rmtree(self.temp_templates_dir)
            shutil.copytree(self.templates_sample_dir, self.temp_templates_dir)
            os.environ["RECAPTCHA_SITE_KEY"] = ""
            os.environ["RECAPTCHA_SECRET_KEY"] = ""
            os.environ["SMTP_HOST"] = ""
            os.environ["SMTP_PORT"] = "0"
            os.environ["SMTP_ADDRESS"] = ""
            os.environ["SMTP_PASSWORD"] = ""
        else:
            load_env_file()

        Config.load_config(self.module_name, "")
        self.app = app
        self.db = db
        self.config = Config
        self.db_path = Config.DB_URI.split("sqlite:///")[1]

        if os.path.isfile(self.db_path):
            os.remove(self.db_path)

        init_flask(
            self.module_name,
            "",
            self.root_path,
            Config,
            self.models,
            self.blueprints
        )
        self.app.app_context().push()

        self.client = self.app.test_client()
        self.context = self.app.test_request_context()

    def tearDown(self):
        """
        Removes any generated files from the filesystem and handles other
        post-test tasks
        :return: None
        """
        self.cleanup()

    def cleanup(self):
        """
        Cleans up the filesystem after/before tests
        :return: None
        """
        if self.db_path != "" and os.path.isfile(self.db_path):
            os.remove(self.db_path)
        if os.path.isdir(self.temp_templates_dir):
            shutil.rmtree(self.temp_templates_dir)

    def generate_sample_user(self, confirmed: bool = True) \
            -> Tuple[User, str, str]:
        """
        Generates a random user for use in tests
        :param confirmed: Whether or not the user should be confirmed
        :return: The User object, the password and the confirmation key
        """
        password = generate_random(20)
        confirm_key = generate_random(20)
        user = User(
            username=generate_random(12),
            password_hash=generate_hash(password),
            email=generate_random(8) + "@example.com",
            confirmed=confirmed,
            confirmation_hash=generate_hash(confirm_key)
        )
        self.db.session.add(user)
        self.db.session.commit()
        return user, password, confirm_key

    def login_user(self, user: User, password: str):
        """
        Logs in a user
        :param user: The user to log in
        :param password: The password to use
        :return: None
        """
        self.client.post("/login", follow_redirects=True, data={
            "username": user.username,
            "password": password
        })

    def generate_api_key(self, user: User) \
            -> Tuple[ApiKey, str, Dict[str, str]]:
        """
        Generates an API key and base64 encoded headers for requests
        :param user: The user for which to create the key
        :return: The API key object, the actual API key, the headers
        """
        key = generate_random(20)
        hashed = generate_hash(key)
        api_key_obj = ApiKey(user=user, key_hash=hashed)
        self.db.session.add(api_key_obj)
        self.db.session.commit()
        api_key = "{}:{}".format(api_key_obj.id, key)

        return api_key_obj, api_key, self.generate_api_key_headers(api_key)

    # noinspection PyMethodMayBeStatic
    def generate_api_key_headers(self, api_key: str) -> Dict[str, str]:
        """
        Creates HTTP Authorization headers for an API key
        :param api_key: The API key to put in the headers
        :return: The headers
        """
        encoded = b64encode(api_key.encode("utf-8")).decode("utf-8")
        return {
            "Authorization": "Basic {}".format(encoded)
        }



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
import sys
import logging
from logging.handlers import RotatingFileHandler
from typing import Callable, Optional, Union, List
from argparse import ArgumentParser, Namespace
from puffotter.os import makedirs


def cli_start(
        main_func: Union[
            Callable[[], None],
            Callable[[Namespace], None],
            Callable[[Namespace, logging.Logger], None]
        ],
        arg_parser: ArgumentParser,
        exit_msg: str = "Goodbye",
        package_name: Optional[str] = None,
        sentry_dsn: Optional[str] = None,
        release_name: Optional[str] = None
):
    """
    Starts a program and sets up logging, as well as sentry error tracking
    :param main_func: The main function to call
    :param arg_parser: The argument parser to use
    :param exit_msg: The message printed when the program's execution is
                     stopped using a keyboard interrupt
    :param package_name: The package name of the application
    :param sentry_dsn: The sentry DSN to use
    :param release_name: The name of the release
    :return: None
    """
    try:
        args = arg_parser.parse_args()
        setup_logging(args, package_name)

        if release_name is None:
            if package_name is not None:
                import pkg_resources
                version = pkg_resources.get_distribution(package_name).version
                release_name = package_name + "-" + version
            else:
                release_name = "Unknown"
                package_name = "unknown"

        if sentry_dsn is not None:
            import sentry_sdk
            sentry_sdk.init(sentry_dsn, release=release_name)

        from inspect import signature
        sign = signature(main_func)
        if len(sign.parameters) == 0:
            main_func()  # type: ignore
        elif len(sign.parameters) == 1:
            main_func(args)  # type: ignore
        elif len(sign.parameters) == 2:
            logger = logging.getLogger(package_name)
            main_func(args, logger)  # type: ignore
        else:
            print("Invalid amount of parameters for main function")
    except KeyboardInterrupt:
        print(exit_msg)


def argparse_add_verbosity(parser: ArgumentParser):
    """
    Adds --quiet, --verbose and --debug parameters to an ArgumentParser
    :param parser: the parser to which to add those flags
    :return: None
    """
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="Sets the verbosity level of the program to "
                             "'quiet'")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="Sets the verbosity level of the program to "
                             "'verbose'")
    parser.add_argument("-d", "--debug", action="store_true",
                        help="Sets the verbosity level of the program to "
                             "'debug'")
    parser.add_argument("--silent", action="store_true",
                        help="Silences all output to STDOUT")


def argparse_add_logfile(parser: ArgumentParser):
    """
    Adds the --logfile argument to the argument parser
    :param parser: The argument parser to modify
    :return: None
    """
    parser.add_argument("--logfile",
                        help="Specifies the location of a logfile")


def setup_logging(args: Namespace, package_name: Optional[str]):
    """
    Sets up logging for the provided arguments
    :param args: The CLI arguments
    :param package_name: The package name
    :return: None
    """

    if "silent" in args and args.silent:
        loglevel = 100  # Disable all logging output
        sys.stdout = open(os.devnull, "w")  # Disable all print output
    elif "quiet" in args and args.quiet:
        loglevel = logging.ERROR
    elif "verbose" in args and args.verbose:
        loglevel = logging.INFO
    elif "debug" in args and args.debug:
        loglevel = logging.DEBUG
    else:
        loglevel = logging.WARNING

    file_formatter = logging.Formatter(
        "[%(asctime)s] %(levelname)s:[%(filename)s:%(funcName)s(%(lineno)d)] "
        "%(message)s",
        datefmt="%Y-%m-%d:%H-%M-%S")

    stream_formatter = logging.Formatter("%(levelname)s: %(message)s")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(stream_formatter)
    stream_handler.setLevel(loglevel)

    handlers = [stream_handler]  # type: List[logging.Handler]

    if "logfile" in args and args.logfile is not None:
        log_file = args.logfile
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(file_formatter)
        file_handler.setLevel(logging.DEBUG)
        handlers.append(file_handler)

    if package_name is not None:
        log_dir = os.path.join(
            os.path.expanduser("~"), ".config", package_name
        )
        makedirs(log_dir)
        log_file = os.path.join(log_dir, "logs.log")
        rotating_handler = RotatingFileHandler(
            log_file, maxBytes=50000000, backupCount=10
        )
        rotating_handler.setFormatter(file_formatter)
        rotating_handler.setLevel(logging.DEBUG)
        handlers.append(rotating_handler)

    logging.basicConfig(level=logging.DEBUG, handlers=handlers)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
import pathlib
from puffotter.flask.test.TestFramework import _TestFramework
from puffotter.flask.Config import Config


class TestConfig(_TestFramework):
    """
    Class that tests the config class
    """

    def test_db_config(self):
        """
        Tests the database configuration
        :return: None
        """
        os.environ.pop("FLASK_TESTING")
        os.environ["DB_MODE"] = "mysql"
        os.environ["MYSQL_USER"] = "abc"
        os.environ["MYSQL_PASSWORD"] = "def"
        os.environ["MYSQL_HOST"] = "ghi"
        os.environ["MYSQL_PORT"] = "1000"
        os.environ["MYSQL_DATABASE"] = "xyz"
        Config.load_config("puffotter", "sentry_dsn")
        self.assertEqual(Config.DB_URI, "mysql://abc:def@ghi:1000/xyz")

    def test_version(self):
        """
        Tests if the version is fetched correctly
        :return: None
        """
        version_file = os.path.join(
            pathlib.Path(__file__).parent.absolute(),
            "../../../../version"
        )
        with open(version_file, "r") as f:
            version = f.read()
        self.assertEqual(version, Config.VERSION)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import json
from puffotter.flask.test.TestFramework import _TestFramework
from puffotter.flask.db.ApiKey import ApiKey


class TestApiKeyRoute(_TestFramework):
    """
    Class that tests API-key related features
    """

    def test_requesting_api_key(self):
        """
        Tests requesting an API key
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        self.assertEqual(len(ApiKey.query.all()), 0)
        resp = self.client.post("/api/v1/key", json={
            "username": user.username,
            "password": password
        })
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.data.decode("utf-8"))

        self.assertEqual(len(ApiKey.query.all()), 1)
        self.assertEqual(data["status"], "ok")

        api_key = data["data"]["api_key"]
        api_key_obj = ApiKey.query.filter_by(user=user).first()
        api_headers = self.generate_api_key_headers(api_key)

        self.assertTrue(api_key_obj.verify_key(api_key))

        resp = self.client.get(
            "/api/v1/authorize", headers=api_headers, json={}
        )
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.data.decode("utf-8"))
        self.assertEqual(data["status"], "ok")

    def test_requesting_invalid_api_keys(self):
        """
        Tests requesting API keys with invalid data
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        user2, password2, _ = self.generate_sample_user(False)
        base = {
            "username": user.username,
            "password": password
        }
        for args in [
            {"username": "ABC", "expected": "user does not exist"},
            {"password": "ABC", "expected": "password is incorrect"},
            {"username": user2.username, "password": password2,
             "expected": "user is not confirmed"}
        ]:
            self.assertEqual(len(ApiKey.query.all()), 0)
            params = dict(base)
            params.update(args)
            resp = self.client.post("/api/v1/key", json=params)
            self.assertEqual(resp.status_code, 401)
            data = json.loads(resp.data.decode("utf-8"))
            self.assertEqual(data["status"], "error")
            self.assertEqual(data["reason"], args["expected"])
            self.assertEqual(len(ApiKey.query.all()), 0)

    def test_revoking_api_key(self):
        """
        Tests revoking an API key
        :return: None
        """
        user, _, _ = self.generate_sample_user()
        _, api_key, _ = self.generate_api_key(user)

        self.assertEqual(len(ApiKey.query.all()), 1)

        resp = self.client.delete("/api/v1/key", json={"api_key": api_key})
        self.assertEqual(resp.status_code, 200)
        data = json.loads(resp.data.decode("utf-8"))
        self.assertEqual(data["status"], "ok")

        self.assertEqual(len(ApiKey.query.all()), 0)

    def test_unsuccessfully_revoking_api_key(self):
        """
        Tests unsuccessfully revoking an API key
        :return: None
        """
        user, _, _ = self.generate_sample_user()
        api_key_obj, api_key, _ = self.generate_api_key(user)

        for params in [
            {"api_key": "ABC", "expected": "API key does not exist"},
            {"api_key": "{}:ABC".format(api_key_obj.id),
             "expected": "API key not valid"}
        ]:
            self.assertEqual(len(ApiKey.query.all()), 1)

            resp = self.client.delete("/api/v1/key", json=params)
            self.assertEqual(resp.status_code, 401)
            data = json.loads(resp.data.decode("utf-8"))
            self.assertEqual(data["status"], "error")
            self.assertEqual(data["reason"], params["expected"])

            self.assertEqual(len(ApiKey.query.all()), 1)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import json
import requests
from typing import Dict, Any, Optional


class GraphQlClient:
    """
    A simple API wrapper for GraphQL APIs
    """

    def __init__(self, api_url: str):
        """
        Initializes the GraphQL API wrapper
        :param api_url: The API endpoint URL
        """
        self.api_url = api_url

    def query(
            self,
            query_string: str,
            variables: Optional[Dict[str, Any]]
    ) -> Optional[Dict[str, Any]]:
        """
        Executes a GraphQL query
        :param query_string: The query string to use
        :param variables: The variables to send
        :return: The response JSON, or None if an error occurred.
        """
        if variables is None:
            variables = {}

        resp = requests.post(self.api_url, json={
            "query": query_string,
            "variables": variables
        })
        if not resp.status_code < 300:
            return None
        else:
            return json.loads(resp.text)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
import pkg_resources
from typing import Type, Dict


class Config:
    """
    Class that keeps track of configuration data
    The class attributes should only be called after running load_config
    """

    @classmethod
    def load_config(cls, module_name: str, sentry_dsn: str):
        """
        Loads the configuration from environment variables
        :param module_name: The name of the project's module
        :param sentry_dsn: The sentry DSN used for error logging
        :return: None
        """
        Config.LOGGING_PATH = os.environ.get(
            "LOGGING_PATH",
            os.path.join("/tmp", f"{module_name}.log")
        )
        Config.SENTRY_DSN = sentry_dsn
        Config.VERSION = pkg_resources.get_distribution(module_name).version
        Config.FLASK_SECRET = os.environ["FLASK_SECRET"]
        Config.TESTING = os.environ.get("FLASK_TESTING") == "1"

        if Config.TESTING:
            Config.DB_MODE = "sqlite"
        else:
            Config.DB_MODE = os.environ["DB_MODE"].lower()
        if Config.DB_MODE == "sqlite":
            sqlite_path = os.environ.get(
                "SQLITE_PATH",
                os.path.join("/tmp", f"{module_name}.db")
            )
            Config.DB_URI = "sqlite:///" + sqlite_path
        else:
            base = Config.DB_MODE.upper() + "_"
            db_host = os.environ[base + "HOST"]
            db_port = os.environ[base + "PORT"]
            db_user = os.environ[base + "USER"]
            db_password = os.environ[base + "PASSWORD"]
            db_database = os.environ[base + "DATABASE"]
            Config.DB_URI = f"{Config.DB_MODE}://{db_user}:{db_password}@" \
                            f"{db_host}:{db_port}/{db_database}"

        Config.RECAPTCHA_SITE_KEY = os.environ["RECAPTCHA_SITE_KEY"]
        Config.RECAPTCHA_SECRET_KEY = os.environ["RECAPTCHA_SECRET_KEY"]

        Config.SMTP_HOST = os.environ["SMTP_HOST"]
        Config.SMTP_PORT = int(os.environ["SMTP_PORT"])
        Config.SMTP_ADDRESS = os.environ["SMTP_ADDRESS"]
        Config.SMTP_PASSWORD = os.environ["SMTP_PASSWORD"]

        Config._load_extras(Config)

    @classmethod
    def _load_extras(cls, parent: Type["Config"]):
        """
        This method can be used to add attributes in subclasses as well as
        change attributes in the base Config class
        :param parent: The base Config class, used to chage attributes
        :return: None
        """
        pass

    VERSION: str
    """
    The current version of the application
    """

    SENTRY_DSN: str
    """
    The sentry DSN used for error logging
    """

    FLASK_SECRET: str
    """
    The flask secret key
    """

    TESTING: bool
    """
    Whether or not testing is enabled
    """

    LOGGING_PATH: str
    """
    The path to the logging file
    """

    FLASK_PORT: int = int(os.environ.get("FLASK_PORT", "8000"))
    """
    The port to use when serving the flask application
    """

    DB_MODE: str
    """
    The database mode (for example 'sqlite' or 'mysql')
    """

    DB_URI: str
    """
    The database URI to use for database operations
    """

    RECAPTCHA_SITE_KEY: str
    """
    Google ReCaptcha site key for bot detection
    """

    RECAPTCHA_SECRET_KEY: str
    """
    Google ReCaptcha secret key for bot detection
    """

    SMTP_HOST: str
    """
    The SMPT Host used for sending emails
    """

    SMTP_PORT: int
    """
    The SMPT Port used for sending emails
    """

    SMTP_ADDRESS: str
    """
    The SMPT Address used for sending emails
    """
    SMTP_PASSWORD: str
    """
    The SMPT Password used for sending emails
    """

    MIN_USERNAME_LENGTH: int = 1
    """
    The Minimum length for usernames
    """

    MAX_USERNAME_LENGTH: int = 12
    """
    The maximum length of usernames
    """

    MAX_API_KEY_AGE: int = 2592000  # 30 days
    """
    The maximum age for API keys
    """

    REQUIRED_TEMPLATES: Dict[str, str] = {
        "index": "static/index.html",
        "about": "static/about.html",
        "privacy": "static/privacy.html",
        "error_page": "static/error_page.html",
        "registration_email": "email/registration.html",
        "forgot_password_email": "email/forgot_password.html",
        "forgot": "user_management/forgot.html",
        "login": "user_management/login.html",
        "profile": "user_management/profile.html",
        "register": "user_management/register.html"
    }
    """
    Specifies required template files
    """

    STRINGS: Dict[str, str] = {
        "401_message": "You are not logged in",
        "500_message": "The server encountered an internal error and "
                       "was unable to complete your request. "
                       "Either the server is overloaded or there "
                       "is an error in the application.",
        "user_does_not_exist": "User does not exist",
        "user_already_logged_in": "User already logged in",
        "user_already_confirmed": "User already confirmed",
        "user_is_not_confirmed": "User is not confirmed",
        "invalid_password": "Invalid Password",
        "logged_in": "Logged in successfully",
        "logged_out": "Logged out",
        "username_length": "Username must be between {} and {} characters "
                           "long",
        "passwords_do_not_match": "Passwords do not match",
        "email_already_in_use": "Email already in use",
        "username_already_exists": "Username already exists",
        "recaptcha_incorrect": "ReCaptcha not solved correctly",
        "registration_successful": "Registered Successfully. Check your email "
                                   "inbox for confirmation",
        "registration_email_title": "Registration",
        "confirmation_key_invalid": "Confirmation key invalid",
        "user_confirmed_successfully": "User confirmed successfully",
        "password_reset_email_title": "Password Reset",
        "password_was_reset": "Password was reset successfully",
        "password_changed": "Password changed successfully",
        "user_was_deleted": "User was deleted"
    }
    """
    Dictionary that defines various strings used in the application.
    Makes it easier to use custom phrases.
    """



import os
from typing import Union
from werkzeug import Response
from flask import Blueprint, redirect, url_for, request, render_template, flash
from flask_login import login_required, current_user, logout_user, login_user
from puffotter.crypto import generate_hash, generate_random
from puffotter.recaptcha import verify_recaptcha
from puffotter.smtp import send_email
from puffotter.flask.base import app, db
from puffotter.flask.Config import Config
from puffotter.flask.db.User import User

user_management_blueprint = Blueprint("user_management", __name__)


@user_management_blueprint.route("/login", methods=["GET", "POST"])
def login() -> Union[Response, str]:
    """
    Page that allows the user to log in
    :return: The response
    """
    if request.method == "POST":
        username = request.form["username"]
        password = request.form["password"]
        remember_me = request.form.get("remember_me") in ["on", True]

        user: User = User.query.filter_by(username=username).first()

        if user is None:
            flash(Config.STRINGS["user_does_not_exist"], "danger")
        elif current_user.is_authenticated:
            flash(Config.STRINGS["user_already_logged_in"], "info")
        elif not user.confirmed:
            flash(Config.STRINGS["user_is_not_confirmed"], "danger")
        elif not user.verify_password(password):
            flash(Config.STRINGS["invalid_password"], "danger")
        else:
            login_user(user, remember=remember_me)
            flash(Config.STRINGS["logged_in"], "success")
            app.logger.info("User {} logged in.".format(current_user.username))
            return redirect(url_for("static.index"))
        return redirect(url_for("user_management.login"))
    else:
        return render_template("user_management/login.html")


@user_management_blueprint.route("/logout", methods=["GET"])
@login_required
def logout() -> Union[Response, str]:
    """
    Logs out the user
    :return: The response
    """
    app.logger.info("User {} logged out.".format(current_user.username))
    logout_user()
    flash(Config.STRINGS["logged_out"], "info")
    return redirect(url_for("static.index"))


@user_management_blueprint.route("/register", methods=["GET", "POST"])
def register() -> Union[Response, str]:
    """
    Page that allows a new user to register
    :return: The response
    """
    if request.method == "POST":
        username = request.form["username"]
        email = request.form["email"]
        password = request.form["password"]
        password_repeat = request.form["password-repeat"]
        recaptcha_result = verify_recaptcha(
            request.remote_addr,
            request.form["g-recaptcha-response"],
            Config.RECAPTCHA_SECRET_KEY
        )

        all_users = User.query.all()
        usernames = [user.username for user in all_users]
        emails = [user.email for user in all_users]

        _min, _max = Config.MIN_USERNAME_LENGTH, Config.MAX_USERNAME_LENGTH

        if len(username) < _min or len(username) > _max:
            flash(Config.STRINGS["username_length"]
                  .format(_min, _max), "danger")
        elif password != password_repeat:
            flash(Config.STRINGS["passwords_do_not_match"], "danger")
        elif username in usernames:
            flash(Config.STRINGS["username_already_exists"], "danger")
        elif email in emails:
            flash(Config.STRINGS["email_already_in_use"], "danger")
        elif not recaptcha_result:
            flash(Config.STRINGS["recaptcha_incorrect"], "danger")
        else:
            confirmation_key = generate_random(32)
            confirmation_hash = generate_hash(confirmation_key)
            user = User(
                username=username,
                email=email,
                password_hash=generate_hash(password),
                confirmation_hash=confirmation_hash
            )
            db.session.add(user)
            db.session.commit()
            email_msg = render_template(
                "email/registration.html",
                host=request.host,
                target=os.path.join(request.host, "confirm"),
                username=username,
                user_id=user.id,
                confirm_key=confirmation_key
            )
            send_email(
                email,
                Config.STRINGS["registration_email_title"],
                email_msg,
                Config.SMTP_HOST,
                Config.SMTP_ADDRESS,
                Config.SMTP_PASSWORD,
                Config.SMTP_PORT
            )
            app.logger.info("User {} registered.".format(user.username))
            flash(Config.STRINGS["registration_successful"], "info")
            return redirect(url_for("static.index"))
        return redirect(url_for("user_management.register"))
    else:
        return render_template("user_management/register.html")


@user_management_blueprint.route("/confirm", methods=["GET"])
def confirm() -> Union[Response, str]:
    """
    Confirms a user
    :return: The response
    """
    user_id = int(request.args["user_id"])
    confirm_key = request.args["confirm_key"]
    user: User = User.query.get(user_id)

    if user is None:
        flash(Config.STRINGS["user_does_not_exist"], "danger")
    elif user.confirmed:
        flash(Config.STRINGS["user_already_confirmed"], "warning")
    elif not user.verify_confirmation(confirm_key):
        flash(Config.STRINGS["confirmation_key_invalid"], "warning")
    else:
        print("D")
        user.confirmed = True
        db.session.commit()
        flash(Config.STRINGS["user_confirmed_successfully"], "success")
    return redirect(url_for("static.index"))


@user_management_blueprint.route("/forgot", methods=["POST", "GET"])
def forgot() -> Union[Response, str]:
    """
    Allows a user to reset their password
    :return: The response
    """
    if request.method == "POST":
        email = request.form["email"]
        recaptcha_result = verify_recaptcha(
            request.remote_addr,
            request.form["g-recaptcha-response"],
            Config.RECAPTCHA_SECRET_KEY
        )
        user: User = User.query.filter_by(email=email).first()

        if not recaptcha_result:
            flash(Config.STRINGS["recaptcha_incorrect"], "danger")
            return redirect(url_for("user_management.forgot"))
        else:
            if user is None:
                # Fail silently to ensure that a potential attacker can't use
                # the response to figure out information on registered users
                pass
            else:
                new_pass = generate_random(20)
                user.password_hash = generate_hash(new_pass)
                db.session.commit()

                email_msg = render_template(
                    "email/forgot_password.html",
                    host=request.host,
                    target=os.path.join(request.host, "login"),
                    password=new_pass,
                    username=user.username
                )
                send_email(
                    email,
                    Config.STRINGS["password_reset_email_title"],
                    email_msg,
                    Config.SMTP_HOST,
                    Config.SMTP_ADDRESS,
                    Config.SMTP_PASSWORD,
                    Config.SMTP_PORT
                )
            flash(Config.STRINGS["password_was_reset"], "success")
            return redirect(url_for("static.index"))

    else:
        return render_template("user_management/forgot.html")


@user_management_blueprint.route("/profile", methods=["GET"])
@login_required
def profile() -> Union[Response, str]:
    """
    Allows a user to edit their profile details
    :return: The response
    """
    return render_template("user_management/profile.html")


@user_management_blueprint.route("/change_password", methods=["POST"])
@login_required
def change_password() -> Union[Response, str]:
    """
    Allows the user to change their password
    :return: The response
    """
    old_password = request.form["old_password"]
    new_password = request.form["new_password"]
    password_repeat = request.form["password_repeat"]
    user: User = current_user

    if new_password != password_repeat:
        flash(Config.STRINGS["passwords_do_not_match"], "danger")
    elif not user.verify_password(old_password):
        flash(Config.STRINGS["invalid_password"], "danger")
    else:
        user.password_hash = generate_hash(new_password)
        db.session.commit()
        flash(Config.STRINGS["password_changed"], "success")
    return redirect(url_for("user_management.profile"))


@user_management_blueprint.route("/delete_user", methods=["POST"])
@login_required
def delete_user() -> Union[Response, str]:
    """
    Allows a user to delete their account
    :return: The response
    """
    password = request.form["password"]
    user: User = current_user

    if not user.verify_password(password):
        flash(Config.STRINGS["invalid_password"], "danger")
    else:
        app.logger.info("Deleting user {}".format(user))
        db.session.delete(user)
        db.session.commit()
        logout_user()
        flash(Config.STRINGS["user_was_deleted"], "success")
        return redirect(url_for("static.index"))
    return redirect(url_for("user_management.profile"))



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os


def load_env_file(path: str = ".env"):
    """
    Loads an environment file into os.environ
    :param path: THe path to the environment file
    :return: None
    """
    if os.path.isfile(path):
        with open(path, "r") as f:
            for line in f.read().split("\n"):
                try:
                    key, value = line.split("#")[0].split("=", 1)
                    os.environ[key] = value
                except ValueError:
                    pass



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fat-ffipd.

fat-ffipd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fat-ffipd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fat-ffipd.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any
from flask import Blueprint, request
from flask_login import login_required
from puffotter.crypto import generate_random, generate_hash
from puffotter.flask.base import db
from puffotter.flask.Config import Config
from puffotter.flask.db.User import User
from puffotter.flask.db.ApiKey import ApiKey
from puffotter.flask.exceptions import ApiException
from puffotter.flask.routes.decorators import api, api_login_required

user_management_api_blueprint = Blueprint("user_management_api", __name__)


@user_management_api_blueprint.route(
    "/api/v1/key", methods=["POST", "DELETE"]
)
@api
def api_key() -> Dict[str, Any]:
    """
    Allows users to request a new API key or revoke an existing API key
    :return: The JSON response
    """
    data = request.get_json()
    if request.method == "POST":
        username = data["username"]
        password = data["password"]
        user = User.query.filter_by(username=username).first()

        if user is None:
            raise ApiException("user does not exist", 401)
        elif not user.confirmed:
            raise ApiException("user is not confirmed", 401)
        elif not user.verify_password(password):
            raise ApiException("password is incorrect", 401)
        else:
            key = generate_random(32)
            hashed = generate_hash(key)
            _api_key = ApiKey(user=user, key_hash=hashed)
            db.session.add(_api_key)
            db.session.commit()

            return {
                "api_key": "{}:{}".format(_api_key.id, key),
                "expiration": (
                        int(_api_key.creation_time)
                        + Config.MAX_API_KEY_AGE
                ),
                "user": user.__json__(True)
            }

    else:  # request.method == "DELETE"
        key = data["api_key"]
        _api_key = ApiKey.query.get(key.split(":", 1)[0])

        if _api_key is None:
            raise ApiException("API key does not exist", 401)
        elif not _api_key.verify_key(key):
            raise ApiException("API key not valid", 401)
        else:
            db.session.delete(_api_key)
            db.session.commit()
            return {}


@user_management_api_blueprint.route("/api/v1/authorize", methods=["GET"])
@api_login_required
@login_required
@api
def api_authorize() -> Dict[str, Any]:
    """
    Allows a user to check if an API key is authorized or not
    :return: None
    """
    return {}  # Checks done by @login_required



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from threading import Thread
from typing import Callable, Tuple, Dict, Type
from cheroot.wsgi import Server, PathInfoDispatcher
from puffotter.flask.base import app
from puffotter.flask.Config import Config


def __start_background_tasks(
        task_definitions: Dict[str, Tuple[int, Callable]]
):
    """
    Starts background tasks for a flask application
    :param task_definitions: The background tasks, consisting of:
                                - the name of the task
                                - the time to wait before running
                                  the task again
                                - a function that takes no arguments that
                                  executes the task
    :return: None
    """
    def task_factory(_name: str, _delay: int, _function: Callable) -> Thread:
        def run_task():
            while True:
                try:
                    with app.app_context():
                        _function()
                except Exception as error:
                    app.logger.error(f"Encountered exception in "
                                     f"background task {_name}: {error}")
                time.sleep(_delay)
        return Thread(target=run_task)

    tasks = []
    for name, (delay, function) in task_definitions.items():
        tasks.append(task_factory(name, delay, function))

    for thread in tasks:
        thread.start()


def start_server(
        config: Type[Config],
        task_definitions: Dict[str, Tuple[int, Callable]]
):
    """
    Starts the flask application using a cheroot WSGI server
    :param config: The configuration to use
    :param task_definitions: The background tasks, consisting of:
                                - the name of the task
                                - the time to wait before running
                                  the task again
                                - a function that takes no arguments that
                                  executes the task
    :return: None
    """
    __start_background_tasks(task_definitions)

    server = Server(
        ("0.0.0.0", config.FLASK_PORT),
        PathInfoDispatcher({"/": app})
    )

    try:
        server.start()
    except KeyboardInterrupt:
        server.stop()



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from unittest.mock import patch
from puffotter.flask.test.TestFramework import _TestFramework
from puffotter.flask.wsgi import start_server, \
    __start_background_tasks as start_background_tasks


class TestWsgi(_TestFramework):
    """
    Tests the WSGI server
    """

    def test_starting_and_stopping_wsgi_server(self):
        """
        Tests starting and stopping the WSGI server
        :return: None
        """
        class Server:
            def __init__(self, *arg, **kwargs):
                pass

            def start(self):
                raise KeyboardInterrupt()

            def stop(self):
                pass

        with patch("puffotter.flask.wsgi.Server", Server):
            start_server(self.config, {})

    def test_starting_background_tasks(self):
        """
        Tests starting background tasks
        :return: None
        """
        self.config.START = time.time()

        def test_task():
            delta = time.time() - self.config.START
            if 1 < delta < 2:
                raise ValueError("TestException")
            elif delta >= 2:
                exit(0)

        tasks = {
            "test_task": (1, test_task)
        }
        start_background_tasks(tasks)
